from typing import Any, Union

import jax
import jax.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.lax as lax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, Bool, Shaped

from ..._filters import combine, is_array, partition
from ..._module import field, Module
from ..._tree import tree_at, tree_equal
from ..._unvmap import unvmap_any
from .._nontraceable import nonbatchable
from .._primitive import create_vprim


def _select_if_vmap_impl(pred, x, y):
    # Not including the following, as it destroys performance on the GPU: it seems like
    # a copy of `x` is being made.
    #
    # msg = (
    #     "Internal error in Equinox. Please report a bug at "
    #     "https://github.com/patrick-kidger/equinox."
    # )
    # x = error_if(x, jnp.invert(pred), msg)
    return x


def _select_if_vmap_abstract(pred, x, y):
    return x


def _select_if_vmap_jvp(primals, tangents):
    pred, x, y = primals
    _, tx, ty = tangents
    assert x.shape == tx.aval.shape
    assert x.dtype == tx.aval.dtype
    assert y.shape == ty.aval.shape
    assert y.dtype == ty.aval.dtype
    out = _select_if_vmap(pred, x, y, makes_false_steps=False)
    if type(tx) is ad.Zero and type(ty) is ad.Zero:
        t_out = tx
    else:
        if type(tx) is ad.Zero:
            tx = jnp.zeros(tx.aval.shape, tx.aval.dtype)  # pyright: ignore
        if type(ty) is ad.Zero:
            ty = jnp.zeros(ty.aval.shape, ty.aval.dtype)  # pyright: ignore
        t_out = _select_if_vmap(pred, tx, ty, makes_false_steps=False)
    return out, t_out


def _select_if_vmap_transpose(ct, pred, x, y):
    assert not ad.is_undefined_primal(pred)
    assert ct.shape == x.aval.shape
    assert ct.dtype == x.aval.dtype
    assert ct.shape == y.aval.shape
    assert ct.dtype == y.aval.dtype
    if type(ct) is ad.Zero:
        ct_x = None
        ct_y = None
    else:
        zero = jnp.zeros(ct.shape, ct.dtype)
        if ad.is_undefined_primal(x):
            ct_x = _select_if_vmap(pred, ct, zero, makes_false_steps=False)
        else:
            ct_x = None
        if ad.is_undefined_primal(y):
            ct_y = _select_if_vmap(pred, zero, ct, makes_false_steps=False)
        else:
            ct_y = None
    return [None, ct_x, ct_y]


def _select_if_vmap_batch(axis_size, axis_name, trace, inputs, batch_axes):
    del axis_name, trace
    pred, x, y = inputs
    bp, bx, by = batch_axes
    if bp is batching.not_mapped:
        if bx is batching.not_mapped:
            x = jnp.broadcast_to(x, (axis_size,) + x.shape)
        else:
            x = jnp.moveaxis(x, bx, 0)
        if by is batching.not_mapped:
            y = jnp.broadcast_to(y, (axis_size,) + y.shape)
        else:
            y = jnp.moveaxis(y, by, 0)
        out = _select_if_vmap(pred, x, y, makes_false_steps=False)
    else:
        out = jax.vmap(lax.select, in_axes=(bp, bx, by))(pred, x, y)
    return out, 0


select_if_vmap_p = jax.core.Primitive("select_if_vmap")
select_if_vmap_p.def_impl(_select_if_vmap_impl)
select_if_vmap_p.def_abstract_eval(_select_if_vmap_abstract)
ad.primitive_jvps[select_if_vmap_p] = _select_if_vmap_jvp
ad.primitive_transposes[select_if_vmap_p] = _select_if_vmap_transpose
batching.axis_primitive_batchers[select_if_vmap_p] = _select_if_vmap_batch
mlir.register_lowering(
    select_if_vmap_p, mlir.lower_fun(_select_if_vmap_impl, multiple_results=False)
)


# We want to insert `lax.select`s to avoid updating any batch elements in the loop that
# have a False predicate. (But the loop is still going whilst other batch elements have
# a True predicate). However, if we have no vmap at all, then we can be slightly more
# efficient: don't introduce a select at all.
def _select_if_vmap(pred, x, y, makes_false_steps):
    """As `lax.select(pred, x, y)` if `pred` is vmap'd. Not-vmap'd `pred` are assumed to
    be `True`, so that in this case `x` is returned unconditionally.
    """
    if makes_false_steps:
        return lax.select(pred, x, y)
    else:
        pred = fixed_asarray(pred)
        assert pred.shape == ()
        assert pred.dtype == jnp.bool_
        x = fixed_asarray(x)
        y = fixed_asarray(y)
        assert x.shape == y.shape
        assert x.dtype == y.dtype
        return select_if_vmap_p.bind(pred, x, y)


def _maybe_set_impl(
    pred, xs, x, *i_dynamic_leaves, i_static, i_treedef, kwargs, makes_false_steps
):
    i = combine(i_static, jtu.tree_unflatten(i_treedef, i_dynamic_leaves))
    x = _select_if_vmap(pred, x, xs.at[i].get(**kwargs), makes_false_steps)
    return [xs.at[i].set(x, **kwargs)]


def _maybe_set_abstract(
    pred, xs, x, *i_dynamic_leaves, i_static, i_treedef, kwargs, makes_false_steps
):
    return [xs]


def _maybe_set_jvp(
    primals, tangents, *, i_static, i_treedef, kwargs, makes_false_steps
):
    pred, xs, x, *i_dynamic_leaves = primals
    _, t_xs, t_x, *_ = tangents
    i = combine(i_static, jtu.tree_unflatten(i_treedef, i_dynamic_leaves))
    out = _maybe_set(pred, xs, x, i, kwargs=kwargs, makes_false_steps=makes_false_steps)
    if type(t_x) is ad.Zero and type(t_xs) is ad.Zero:
        t_out = t_xs
    else:
        if type(t_x) is ad.Zero:
            t_x = jnp.zeros(t_x.aval.shape, t_x.aval.dtype)  # pyright: ignore
        if type(t_xs) is ad.Zero:
            t_xs = jnp.zeros(t_xs.aval.shape, t_xs.aval.dtype)  # pyright: ignore
        t_out = _maybe_set(
            pred, t_xs, t_x, i, kwargs=kwargs, makes_false_steps=makes_false_steps
        )
    return [out], [t_out]


def _maybe_set_transpose(
    ct_out,
    pred,
    xs,
    x,
    *i_dynamic_leaves,
    i_static,
    i_treedef,
    kwargs,
    makes_false_steps
):
    assert not ad.is_undefined_primal(pred)
    for z in i_dynamic_leaves:
        assert not ad.is_undefined_primal(z)
    i = combine(i_static, jtu.tree_unflatten(i_treedef, i_dynamic_leaves))
    [ct_out] = ct_out
    if ad.is_undefined_primal(xs):
        # Not updating a zero! `_Buffer`s are write-once to each location, so when
        # transposed they are read-once from each location, meaning if we read from `i`
        # now then we will never do so again, so we can skip placing a zero there.
        #
        # Besides just being a nice efficiency gain: this is actually important for
        # working around b/288798733, in which having extra dynamic_update_slices causes
        # operations to happen out-of-place.
        ct_xs = ct_out
    else:
        ct_xs = None
    if ad.is_undefined_primal(x):
        if type(ct_out) is ad.Zero:
            ct_x = None
        else:
            ct_x = ct_out.at[i].get(**kwargs)
            ct_x = _select_if_vmap(pred, ct_x, jnp.zeros_like(ct_x), makes_false_steps)
    else:
        ct_x = None
    return [None, ct_xs, ct_x, None]


maybe_set_p = create_vprim(
    "maybe_set",
    _maybe_set_impl,
    _maybe_set_abstract,
    _maybe_set_jvp,
    _maybe_set_transpose,
)


# This is a carefully optimised routine, that relies on the special behaviour exhibited
# by `_Buffer`.
# The main one is fact that `_Buffer`s are write-once to each location, so they
# are read-once to each location when transposed. This means we can skip placing zeros
# in the cotangent of `xs`. This reduces the number of in-place updates on the backward
# pass from 2 to 0. This ends up being really important for efficiency, due to
# b/288798733.
# Second, the fact that unbatched `pred` are necessarily always True (due to being
# used inside a while loop) means that we use `_select_if_vmap` over simply
# `lax.select`.
def _maybe_set(pred, xs, x, i, *, kwargs, makes_false_steps):
    """As `lax.select(pred, xs.at[i].set(x, **kwargs), xs)`, under the assumption that
    every location `i` is written to at most once. (So that we can have a more efficient
    transpose rule. Also assumes that non-vmap'd `pred` is always `True`.)
    """
    if jnp.shape(pred) != () or jnp.result_type(pred) != jnp.bool_:
        raise ValueError("predicate must be a boolean scalar.")
    dtype = jnp.result_type(x, xs)
    if dtype != jnp.result_type(xs):
        raise ValueError(
            "When doing `buffer.at[i].set(value)`, then `value` must have a dtype that "
            "can be promoted to the same dtype as `buffer`."
        )
    x = fixed_asarray(x).astype(dtype)
    x = jnp.broadcast_to(x, jax.eval_shape(lambda: xs[i]).shape)
    i_dynamic, i_static = partition(i, is_array)
    i_dynamic_leaves, i_treedef = jtu.tree_flatten(i_dynamic)
    [out] = maybe_set_p.bind(
        pred,
        xs,
        x,
        *i_dynamic_leaves,
        i_static=i_static,
        i_treedef=i_treedef,
        kwargs=kwargs,
        makes_false_steps=makes_false_steps
    )
    return out


class _Buffer(Module):
    _array: Union[Shaped[Array, "..."], "_Buffer"]
    _pred: Bool[Array, ""]
    _tag: object = field(static=True)
    _makes_false_steps: bool = field(static=True)

    def __getitem__(self, item):
        return self._array[item]

    def _op(self, pred, item, x, op, kwargs, makes_false_steps):
        pred = pred & self._pred
        if isinstance(self._array, _Buffer):
            array = self._array._op(pred, item, x, op, kwargs, makes_false_steps)
        else:
            array = op(
                pred,
                self._array,
                x,
                item,
                kwargs=kwargs,
                makes_false_steps=makes_false_steps,
            )
        return _Buffer(array, self._pred, self._tag, self._makes_false_steps)

    @property
    def at(self):
        return _BufferAt(self, self._makes_false_steps)

    @property
    def shape(self):
        return self._array.shape

    @property
    def dtype(self):
        return self._array.dtype

    @property
    def size(self):
        return self._array.size


class _BufferAt(Module):
    _buffer: _Buffer
    _makes_false_steps: bool = field(static=True)

    def __getitem__(self, item):
        return _BufferItem(self._buffer, item, self._makes_false_steps)


class _BufferItem(Module):
    _buffer: _Buffer
    _item: Any
    _makes_false_steps: bool = field(static=True)

    def set(self, x, *, pred=True, **kwargs):
        if pred is True:
            makes_false_steps = self._makes_false_steps
        else:
            makes_false_steps = True
        return self._buffer._op(
            pred, self._item, x, _maybe_set, kwargs, makes_false_steps
        )


def _is_buffer(x):
    return isinstance(x, _Buffer)


def _unwrap_buffers(x):
    while _is_buffer(x):
        x = x._array
    return x


# Work around JAX issue #15676
@jax.custom_jvp
def fixed_asarray(x):
    return jnp.asarray(x)


@fixed_asarray.defjvp
def _fixed_asarray_jvp(x, tx):
    (x,) = x
    (tx,) = tx
    return fixed_asarray(x), fixed_asarray(tx)


def common_rewrite(cond_fun, body_fun, init_val, max_steps, buffers, makes_false_steps):
    """Handles:

    - Efficient in-place updates;
    - Efficient/correct vmap;
    - max_steps.

    The efficient in-place updates are done using buffers. These mean that we get to
    interchange `select` (from vmap) and `scatter` (from the in-place update) so that
    they happen in the efficient order. Specifically for `checkpointed_while_loop`, it
    also means we don't need to save our buffers in the list of checkpoints.

    The efficient vmap is done by always having an unvmap'd return value from
    `cond_fun`: essentially we reduce from the vmap'd case to the no-vmap case here.
    This ensures that (a) `bounded_while_loop` exhibits early-exit, as our `lax.cond`s
    don't get turned into `lax.select`s, and (b) `checkpointed_while_loop` has the
    correct logic on the backward pass. (As all checkpoints need to be in lockstep
    across the batch.)
    """

    if buffers is None:
        new_buffers = lambda is_leaf: lambda val: []
    else:

        def new_buffers(is_leaf):
            def new_buffers2(val):
                _, _, _, val = val  # ignore step and pred
                bufs = buffers(val)
                # Ignore the ._pred attribute of nested buffers.
                # This is kind of a hack: we're special-casing support for nested
                # buffers so that end users don't have to do the ._array lookup
                # themselves.
                tree = jtu.tree_map(_unwrap_buffers, bufs, is_leaf=_is_buffer)
                return jtu.tree_leaves(tree, is_leaf=is_leaf)

            return new_buffers2

    def _wrap_buffers(val, pred, tag):
        def wrap_buffer(leaf):
            if not is_array(leaf):
                raise ValueError("Only arrays can be treated as buffers.")
            return _Buffer(leaf, pred, tag, makes_false_steps)

        _, _, _, buffer_val = tree_at(
            new_buffers(None), (None, None, None, val), replace_fn=wrap_buffer
        )
        return buffer_val

    def new_cond_fun(val):
        step, pred, prev_pred, val = val
        del pred
        # Note that this is actually recomputing `pred`!
        # For some reason this is actually a minor performance optimisation.
        # See https://github.com/patrick-kidger/diffrax/issues/274
        buffer_val = _wrap_buffers(val, prev_pred, None)
        out = unvmap_any(cond_fun(buffer_val))
        if max_steps is not None:
            if type(max_steps) is not int:
                raise ValueError("`max_steps` must be a Python integer")
            out = out & (step < max_steps)
        return nonbatchable(out)

    def new_body_fun(val):
        tag = object()

        def is_our_buffer(node):
            return isinstance(node, _Buffer) and node._tag is tag

        def unwrap_and_select(leaf, leaf2):
            if is_our_buffer(leaf):
                assert is_our_buffer(leaf2)
                # sanity check that this is the lowest buffer, i.e. when nesting
                # multiple checkpointed_while_loops.
                assert is_array(leaf._array)
                assert is_array(leaf2._array)
                return leaf2._array
            else:
                return _select_if_vmap(pred, leaf2, leaf, makes_false_steps)

        step, pred, _, val = val
        buffer_val = _wrap_buffers(val, pred, tag)
        buffer_val2 = body_fun(buffer_val)
        # Strip `.named_shape`; c.f. Diffrax issue #246
        struct = jax.eval_shape(lambda: buffer_val)
        struct2 = jax.eval_shape(lambda: buffer_val2)
        struct = jtu.tree_map(lambda x: (x.shape, x.dtype), struct)
        struct2 = jtu.tree_map(lambda x: (x.shape, x.dtype), struct2)
        if not tree_equal(struct, struct2):
            raise ValueError("`body_fun` must have the same input and output structure")
        val2 = jtu.tree_map(
            unwrap_and_select, buffer_val, buffer_val2, is_leaf=is_our_buffer
        )
        step2 = step + 1
        pred2 = pred & cond_fun(buffer_val2)
        return step2, pred2, pred, val2

    init_val = jtu.tree_map(fixed_asarray, init_val)
    init_buffer_val = _wrap_buffers(init_val, jnp.array(True), None)
    init_pred = jnp.array(cond_fun(init_buffer_val))
    new_init_val = (jnp.array(0), init_pred, jnp.array(True), init_val)

    return new_cond_fun, new_body_fun, new_init_val, new_buffers



import dataclasses
import functools as ft
import inspect
import weakref
from collections.abc import Callable
from typing import Any, cast, TypeVar, Union
from typing_extensions import dataclass_transform, ParamSpec

import jax.tree_util as jtu
import numpy as np
from jaxtyping import Array, Bool

from ._better_abstract import ABCMeta, dataclass
from ._caches import internal_lru_caches
from ._doc_utils import doc_repr
from ._pretty_print import tree_pformat
from ._tree import tree_equal


_P = ParamSpec("_P")
_T = TypeVar("_T")


def static_field(**kwargs):
    """Deprecated in favour of [`equinox.field`][], i.e. `eqx.field(static=True)`."""
    return field(**kwargs, static=True)


_identity = doc_repr(lambda x: x, "lambda x: x")


def field(
    *, converter: Callable[[Any], Any] = _identity, static: bool = False, **kwargs
):
    """Equinox supports extra functionality on top of the default dataclasses.

    **Arguments:**

    - `converter`: a function to call on this field when the model is initialised. For
        example, `field(converter=jax.numpy.asarray)` to convert
        `bool`/`int`/`float`/`complex` values to JAX arrays.
    - `static`: whether the field should not interact with any JAX transform at all (by
        making it part of the PyTree structure rather than a leaf).
    - `**kwargs`: All other keyword arguments are passed on to `dataclass.field`.

    **Converter**

    !!! example

        ```python
        class MyModule(eqx.Module):
            foo: Array = eqx.field(converter=jax.numpy.asarray)

        mymodule = MyModule(1.0)
        assert isinstance(mymodule.foo, jax.Array)
        ```

    **Static**

    !!! example

        ```python
        class MyModule(eqx.Module):
            normal_field: int
            static_field: int = eqx.field(static=True)

        mymodule = MyModule("normal", "static")
        leaves, treedef = jax.tree_util.tree_flatten(mymodule)
        assert leaves == ["normal"]
        assert "static" in str(treedef)
        ```

    This means that it does not interact with any JAX transforms, like JIT or grad.
    This means that it's usually a bug to make JAX arrays be static fields.

    This is an advanced feature that should very rarely be used. It is preferred to
    just filter out each field with `eqx.partition` whenever you need to select only
    some fields.
    """
    try:
        metadata = dict(kwargs["metadata"])
    except KeyError:
        metadata = kwargs["metadata"] = {}
    if "converter" in metadata:
        raise ValueError("Cannot use metadata with `static` already set.")
    if "static" in metadata:
        raise ValueError("Cannot use metadata with `static` already set.")
    metadata["converter"] = converter
    if static:
        metadata["static"] = True
    return dataclasses.field(**kwargs)


class _wrap_method:
    def __init__(self, method):
        self.method = method
        if getattr(self.method, "__isabstractmethod__", False):
            self.__isabstractmethod__ = self.method.__isabstractmethod__

    def __get__(self, instance, owner):
        if instance is None:
            return self.method
        _method = ft.wraps(self.method)(jtu.Partial(self.method, instance))
        delattr(_method, "__wrapped__")
        return _method


def _not_magic(k: str) -> bool:
    return not (k.startswith("__") and k.endswith("__"))


_has_dataclass_init = weakref.WeakKeyDictionary()


# Inherits from ABCMeta as a convenience for a common use-case.
# It's not a feature we use ourselves.
@dataclass_transform(field_specifiers=(dataclasses.field, field, static_field))
class _ModuleMeta(ABCMeta):
    def __new__(mcs, name, bases, dict_):  # pyright: ignore
        dict_ = {
            k: _wrap_method(v) if _not_magic(k) and inspect.isfunction(v) else v
            for k, v in dict_.items()
        }
        cls = super().__new__(mcs, name, bases, dict_)
        # Do override subclasses' dataclass-__init__-s. (None of which call super, so
        # they must be overriden.)
        # Don't override custom __init__'s, which leads to poor ergonomics:
        # e.g. if `B` has a custom init then `class A(B): pass` would otherwise set a
        # dataclass init that overrides the custom __init__.
        if "__init__" in cls.__dict__:
            _init = False
        else:
            for kls in cls.__mro__:
                try:
                    _init = _has_dataclass_init[kls]
                except KeyError:
                    pass
                else:
                    break
            else:
                assert name == "Module"
                _init = True  # eqx.Module itself
        if _init:
            init_doc = cls.__init__.__doc__
        cls = dataclass(eq=False, repr=False, frozen=True, init=_init)(
            cls  # pyright: ignore
        )
        # must happen after `dataclass(...)` we use this in `__getattribute__` to avoid
        # making any property(def __wrapped__) visible until then. We want to be able to
        # support property(def __wrapped__) for the sake of classes whose instances are
        # wrappers (i.e. via `module_update_wrapper`), but this means that a
        # `__wrapped__` object is also visible on the class object itself, despite not
        # pointing to a callable. `dataclass(...)` spots this and tries to use it form
        # a signature.
        _has_dataclass_init[cls] = _init
        if _init:
            cls.__init__.__doc__ = init_doc  # pyright: ignore
            cls.__init__.__module__ = cls.__module__
        jtu.register_pytree_with_keys(
            cls,
            flatten_with_keys=ft.partial(
                _flatten_module, with_keys=True
            ),  # pyright: ignore
            flatten_func=ft.partial(
                _flatten_module, with_keys=False
            ),  # pyright: ignore
            unflatten_func=ft.partial(_unflatten_module, cls),  # pyright: ignore
        )
        return cls

    # See note above for `_has_dataclass_init`.
    def __getattribute__(cls, item):
        value = super().__getattribute__(item)
        if (
            item == "__wrapped__"
            and isinstance(value, property)
            and cls not in _has_dataclass_init
        ):
            raise AttributeError
        else:
            return value

    def __call__(cls, *args, **kwargs):
        # Defreeze it during __init__
        initable_cls = _make_initable(cls, wraps=False)
        self = super(_ModuleMeta, initable_cls).__call__(*args, **kwargs)
        missing_names = {
            field.name
            for field in dataclasses.fields(cls)  # pyright: ignore
            if field.init and field.name not in dir(self)
        }
        if len(missing_names):
            raise ValueError(
                f"The following fields were not initialised during __init__: "
                f"{missing_names}"
            )
        for field in dataclasses.fields(self):
            try:
                converter = field.metadata["converter"]
            except KeyError:
                pass
            else:
                setattr(self, field.name, converter(getattr(self, field.name)))
        object.__setattr__(self, "__class__", cls)
        return self


_wrapper_field_names = {
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
}


@ft.lru_cache(maxsize=128)
def _make_initable(cls: _ModuleMeta, wraps: bool) -> _ModuleMeta:
    if wraps:
        field_names = _wrapper_field_names
    else:
        field_names = {
            field.name for field in dataclasses.fields(cls)  # pyright: ignore
        }

    class _InitableModule(cls):  # pyright: ignore
        pass

    # Done like this to avoid dataclasses complaining about overriding setattr on a
    # frozen class.
    def __setattr__(self, name, value):
        if name in field_names:
            object.__setattr__(self, name, value)
        else:
            raise AttributeError(f"Cannot set attribute {name}")

    _InitableModule.__setattr__ = __setattr__
    # Make beartype happy
    _InitableModule.__init__ = cls.__init__  # pyright: ignore
    # Appears in AbstractVar error messages
    _InitableModule.__name__ = cls.__name__
    _InitableModule.__qualname__ = cls.__qualname__

    return _InitableModule


internal_lru_caches.append(_make_initable)


# This class exists primarily just to hide the wrapper fields from the PyTreeDef repr.
# Stuff like the docstring can be pretty long and entirely unhelpful to print out.
@dataclass()
class _FlattenedData:
    dynamic_field_names: tuple
    static_field_names: tuple
    static_field_values: tuple
    wrapper_field_names: tuple
    wrapper_field_values: tuple

    def __repr__(self):
        x = (
            self.dynamic_field_names,
            self.static_field_names,
            self.static_field_values,
        )
        return repr(x)[1:-1]


def _flatten_module(module: "Module", with_keys: bool):
    dynamic_field_names = []
    dynamic_field_values = []
    static_field_names = []
    static_field_values = []
    wrapper_field_names = []
    wrapper_field_values = []
    for field_ in dataclasses.fields(module):
        name = field_.name
        try:
            value = module.__dict__[name]
        except KeyError:
            continue
        if field_.metadata.get("static", False):
            static_field_names.append(name)
            static_field_values.append(value)
        else:
            dynamic_field_names.append(name)
            if with_keys:
                dynamic_field_values.append((jtu.GetAttrKey(name), value))
            else:
                dynamic_field_values.append(value)
    sentinel = object()
    for name in _wrapper_field_names:
        value = getattr(module, name, sentinel)
        if value is not sentinel:
            wrapper_field_names.append(name)
            wrapper_field_values.append(value)
    aux = _FlattenedData(
        tuple(dynamic_field_names),
        tuple(static_field_names),
        tuple(static_field_values),
        tuple(wrapper_field_names),
        tuple(wrapper_field_values),
    )
    return tuple(dynamic_field_values), aux


def _unflatten_module(cls: type["Module"], aux: _FlattenedData, dynamic_field_values):
    module = object.__new__(cls)
    for name, value in zip(aux.dynamic_field_names, dynamic_field_values):
        object.__setattr__(module, name, value)
    for name, value in zip(aux.static_field_names, aux.static_field_values):
        object.__setattr__(module, name, value)
    for name, value in zip(aux.wrapper_field_names, aux.wrapper_field_values):
        object.__setattr__(module, name, value)
    return module


class Module(metaclass=_ModuleMeta):
    """Base class. Create your model by inheriting from this.

    This will make your model a
    [dataclass](https://docs.python.org/3/library/dataclasses.html) and a
    [pytree](https://jax.readthedocs.io/en/latest/pytrees.html).

    **Fields**

    Specify all its fields at the class level (identical to
    [dataclasses](https://docs.python.org/3/library/dataclasses.html)). This defines
    its children as a PyTree.

    ```python
    class MyModule(equinox.Module):
        weight: jax.Array
        bias: jax.Array
        submodule: equinox.Module
    ```

    **Initialisation**

    A default `__init__` is automatically provided, which just fills in fields with the
    arguments passed. For example `MyModule(weight, bias, submodule)`.

    Alternatively (quite commonly) you can provide an `__init__` method yourself:

    ```python
    class MyModule(equinox.Module):
        weight: jax.Array
        bias: jax.Array
        submodule: equinox.Module

        def __init__(self, in_size, out_size, key):
            wkey, bkey, skey = jax.random.split(key, 3)
            self.weight = jax.random.normal(wkey, (out_size, in_size))
            self.bias = jax.random.normal(bkey, (out_size,))
            self.submodule = equinox.nn.Linear(in_size, out_size, key=skey)
    ```

    **Methods**

    It is common to create some methods on the class -- for example to define the
    forward pass of a model.

    ```python
    class MyModule(equinox.Module):
        ...  # as above

        def __call__(self, x):
            return self.submodule(x) + self.weight @ x + self.bias
    ```

    !!! tip

        You don't have to define `__call__`:

        - You can define other methods if you want.
        - You can define multiple methods if you want.
        - You can define no methods if you want. (And just use `equinox.Module` as a
            nice syntax for custom PyTrees.)

        No method is special-cased.

    **Usage**

    After you have defined your model, then you can use it just like any other PyTree
    -- that just happens to have some methods attached. In particular you can pass it
    around across `jax.jit`, `jax.grad` etc. in exactly the way that you're used to.

    !!! example

        If you wanted to, then it would be completely safe to do

        ```python
        class MyModule(equinox.Module):
            ...

            @jax.jit
            def __call__(self, x):
                ...
        ```

        because `self` is just a PyTree. Unlike most other neural network libaries, you
        can mix Equinox and native JAX without any difficulties at all.

    !!! tip

        Equinox modules are all [ABCs](https://docs.python.org/3/library/abc.html) by
        default. This means you can use [`abc.abstractmethod`](https://docs.python.org/3/library/abc.html#abc.abstractmethod).
        You can also create abstract instance attributes or abstract class attributes,
        see [`equinox.AbstractVar`][] and [`equinox.AbstractClassVar`][].
    """  # noqa: E501

    def __hash__(self):
        return hash(tuple(jtu.tree_leaves(self)))

    def __eq__(  # pyright: ignore
        self, other
    ) -> Union[bool, np.bool_, Bool[Array, ""]]:
        return tree_equal(self, other)

    def __repr__(self):
        return tree_pformat(self)


# Modifies in-place, just like functools.update_wrapper
def module_update_wrapper(
    wrapper: Module, wrapped: Callable[_P, _T]
) -> Callable[_P, _T]:
    """Like `functools.update_wrapper` (or its better-known cousin, `functools.wraps`),
    but can be used on [`equinox.Module`][]s. (Which are normally immutable.)

    !!! Example

        ```python
        class Wrapper(eqx.Module):
            fn: Callable

            def __call__(self, *args, **kwargs):
                return self.fn(*args, **kwargs)

            @property
            def __wrapped__(self):
                return self.fn

        def make_wrapper(fn):
            return eqx.module_update_wrapper(Wrapper(fn), fn)
        ```

    For example, [`equinox.filter_jit`][] returns a module representing the JIT'd
    computation. `module_update_wrapper` is used on this module to indicate that this
    JIT'd computation wraps the original one. (Just like how `functools.wraps` is used.)

    Note that as in the above example, the wrapper class must supply a `__wrapped__`
    property, which redirects to the wrapped object.
    """
    cls = wrapper.__class__
    if not isinstance(getattr(cls, "__wrapped__", None), property):
        raise ValueError("Wrapper module must supply `__wrapped__` as a property.")
    initable_cls = _make_initable(cls, wraps=True)
    object.__setattr__(wrapper, "__class__", initable_cls)
    try:
        # Like `ft.update_wrapper(wrapper, wrapped, updated=())`.
        # We don't update __dict__ as it's common/possible for wrapper and wrapped to
        # both be classes implementing __call__, in which case copying __dict__ over
        # basically just breaks the wrapper class.
        # We don't set __wrapped__, and instead demand that the wrapper class tell us
        # how to redirect to the wrapped object. This is avoid duplicating part of the
        # PyTree.
        for attr in _wrapper_field_names:
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapper, attr, value)
    finally:
        object.__setattr__(wrapper, "__class__", cls)
    return cast(Callable[_P, _T], wrapper)


class Static(Module):
    value: Any = field(static=True)


class Partial(Module):
    """Like `functools.partial`, but treats the wrapped function, and partially-applied
    args and kwargs, as a PyTree.

    This is very much like `jax.tree_util.Partial`. The difference is that the JAX
    version requires that `func` be specifically a *function* -- and will silently
    misbehave if given any non-function callable, e.g. [`equinox.nn.MLP`][]. In contrast
    the Equinox version allows for arbitrary callables.
    """

    func: Callable
    args: tuple[Any, ...]
    keywords: dict[str, Any]

    def __init__(self, func, /, *args, **kwargs):
        """**Arguments:**

        - `func`: the callable to partially apply.
        - `*args`: any positional arguments to apply.
        - `**kwargs`: any keyword arguments to apply.
        """
        self.func = func
        self.args = args
        self.keywords = kwargs

    def __call__(self, *args, **kwargs):
        """Call the wrapped `self.func`.

        **Arguments:**

        - `*args`: the arguments to apply. Passed after those arguments passed during
            `__init__`.
        - `**kwargs`: any keyword arguments to apply.

        **Returns:**

        The result of the wrapped function.
        """
        return self.func(*self.args, *args, **kwargs, **self.keywords)



from collections.abc import Callable, Hashable
from typing import Any

import jax
import jax.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, PyTree

from .._ad import filter_custom_vjp
from .._doc_utils import WithRepr
from .._errors import error_if
from .._filters import combine, filter, is_array, is_array_like, partition
from .._module import Module
from .._pretty_print import pformat_short_array_text, tree_pformat, tree_pprint


def announce_transform(
    x, name=None, intermediates=False, announce: Callable[[str], Any] = print
):
    """Identity function on an arbitrary PyTree. Announces each time it is parsed as
    part of a jaxpr.
    """
    if name is None:
        name = ""
    else:
        name = f"({name}) "
    array, nonarray = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(array)
    flat = announce_jaxpr_p.bind(
        *flat, stack=(), name=name, intermediates=intermediates, announce=announce
    )
    array = jtu.tree_unflatten(treedef, flat)
    return combine(array, nonarray)


def _impl(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("impl,")
    announce(name + ":".join(stack))
    return x


def _abstract(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("abstract",)
    announce(name + ":".join(stack))
    return x


def _jvp(p, t, *, stack, name, intermediates, announce):
    if intermediates:
        _stack = stack + ("jvp",)
        announce(name + ":".join(_stack))
    p_stack = stack + ("jvp_p",)
    t_stack = stack + ("jvp_t",)
    p_out = announce_jaxpr_p.bind(
        *p, stack=p_stack, name=name, intermediates=intermediates, announce=announce
    )
    t_out = announce_jaxpr_p.bind(
        *t, stack=t_stack, name=name, intermediates=intermediates, announce=announce
    )
    return p_out, t_out


def _transpose(ct, p, *, stack, name, intermediates, announce):
    stack = stack + ("transpose",)
    if intermediates:
        announce(name + ":".join(stack))
    return announce_jaxpr_p.bind(
        *ct, stack=stack, name=name, intermediates=intermediates, announce=announce
    )


def _batching(p, b, *, stack, name, intermediates, announce):
    stack = stack + ("vmap",)
    if intermediates:
        announce(name + ":".join(stack))
    out = announce_jaxpr_p.bind(
        *p, stack=stack, name=name, intermediates=intermediates, announce=announce
    )
    return out, b


def _mlir(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("mlir",)
    announce(name + ":".join(stack))
    return x


announce_jaxpr_p = jax.core.Primitive("announce_jaxpr")
announce_jaxpr_p.multiple_results = True
announce_jaxpr_p.def_impl(_impl)
announce_jaxpr_p.def_abstract_eval(_abstract)
ad.primitive_jvps[announce_jaxpr_p] = _jvp
ad.primitive_transposes[announce_jaxpr_p] = _transpose
batching.primitive_batchers[announce_jaxpr_p] = _batching
mlir.register_lowering(announce_jaxpr_p, mlir.lower_fun(_mlir, multiple_results=True))


def debug_backward_nan(x, name=None, terminate=True):
    """Debug NaNs that only occur on the backward pass.

    **Arguments:**

    - `x`: a variable to intercept.
    - `name`: an optional name to appear in printed debug statements.
    - `terminate`: whether to halt the computation if a NaN cotangent is found.

    **Returns:**

    The `x` argument is returned unchanged.

    As a side-effect, both the primal and the cotangent for `x` will be printed out
    during the backward pass.
    """
    return _debug_backward_nan(x, name, terminate)


@filter_custom_vjp
def _debug_backward_nan(x, name, terminate):
    return x


@_debug_backward_nan.def_fwd
def _debug_backward_nan_fwd(perturbed, x, name, terminate):
    del perturbed
    return debug_backward_nan(x, name, terminate), None


class _LongRepr(Module):
    obj: Any

    def __repr__(self):
        return tree_pformat(self.obj, short_arrays=False)


@_debug_backward_nan.def_bwd
def _debug_backward_nan_bwd(residuals, grad_x, perturbed, x, name, terminate):
    del residuals, perturbed
    msg = "   primals={x}\ncotangents={grad_x}"
    if name is not None:
        msg = f"{name}:\n" + msg
    jax.debug.print(  # pyright: ignore
        msg, x=_LongRepr(x), grad_x=_LongRepr(grad_x), ordered=True
    )
    if terminate:
        nans = [
            jnp.isnan(a).any() for a in jtu.tree_leaves(filter(grad_x, is_array_like))
        ]
        grad_x = error_if(grad_x, jnp.any(jnp.stack(nans)), "Encountered NaN")
    return grad_x


_dce_store = {}


def _register_alive(name: Hashable, tag: object):
    def _register_alive_impl(i, x):
        leaves, _ = _dce_store[name][tag]
        leaves[i.item()] = (x.shape, x.dtype.name)
        return x

    return _register_alive_impl


def store_dce(x: PyTree[Array], name: Hashable = None):
    """Used to check whether an array (or pytree of arrays) is DCE'd.

    `store_dce` must be used within a JIT'd function, and acts as the identity
    function. When the JIT'd function is called, then whether each array got DCE'd or
    not is recorded. This can subsequently be inspected using `inspect_dce`.

    !!! Example:

        ```python
        @jax.jit
        def f(x):
            a, _ = eqxi.store_dce((x**2, x + 1))
            return a

        f(1)
        eqxi.inspect_dce()
        # Found 1 call to `equinox.internal.register_dce`.
        # Entry 0:
        # (i32[], <DCE'd>)
        ```

    **Arguments:**

    - `x`: Any PyTree of JAX arrays.
    - `name`: Optional argument. Any hashable value used to distinguish this call site
        from another call site. If used, then it should be passed to `inspect_dce` to
        print only those entries with this name.

    **Returns:**

    `x` is returned unchanged.
    """
    if not isinstance(jnp.array(1) + 1, jax.core.Tracer):
        raise RuntimeError(
            "`equinox.internal.store_dce` should be used insnide of JIT."
        )
    tag = object()
    leaves, treedef = jtu.tree_flatten(x)
    try:
        _tag_store = _dce_store[name]
    except KeyError:
        _tag_store = _dce_store[name] = {}
    _tag_store[tag] = ({}, treedef)
    leaves = [
        jax.pure_callback(  # pyright: ignore
            _register_alive(name, tag), x, i, x, vectorized=True
        )
        for i, x in enumerate(leaves)
    ]
    return jtu.tree_unflatten(treedef, leaves)


def inspect_dce(name: Hashable = None):
    """Used in conjunction with `equinox.internal.check_dce`; see documentation there.

    Must be called outside of any JIT'd function.

    **Arguments:**

    - `name`: Optional argument. Whatever name was used with `check_dce`.

    **Returns:**

    Nothing. DCE information is printed to stdout.
    """
    if isinstance(jnp.array(1) + 1, jax.core.Tracer):
        raise RuntimeError(
            "`equinox.internal.inspect_dce` should be used outside of JIT."
        )
    try:
        _tag_store = _dce_store[name]
    except KeyError as e:
        raise ValueError(
            "`equinox.internal.inspect_dce` should be called after "
            "`equinox.internal.store_dce` has run."
        ) from e
    new_leaves = []
    maybe_s = "" if len(_tag_store) == 1 else "s"
    print(f"Found {len(_tag_store)} call{maybe_s} to `equinox.internal.register_dce`.")
    for i, (leaves, treedef) in enumerate(_tag_store.values()):
        for j in range(treedef.num_leaves):
            try:
                shape, dtype = leaves[j]
            except KeyError:
                value = "<DCE'd>"
            else:
                value = pformat_short_array_text(shape, dtype)
            new_leaves.append(WithRepr(None, value))
        tree = jtu.tree_unflatten(treedef, new_leaves)
        print(f"Entry {i}:")
        tree_pprint(tree)
